from .build_error_position import build_error_position
from .errors_types import ValidInput
from .get_input_as_string import get_input_as_string

INPUT_PARSER_ERROR_HEADER_MESSAGE = "Failed to parse input string:"


class InputParseError(Exception):
    name = "InputParseError"
    most_recent_input: ValidInput
    pos: int
    previous_input: ValidInput

    def __init__(
        self, most_recent_input: ValidInput, pos: int, previous_input: ValidInput = "",
    ):
        super().__init__(
            "\n".join(
                [
                    INPUT_PARSER_ERROR_HEADER_MESSAGE,
                    "",
                    *build_error_position(
                        f"{get_input_as_string(previous_input)}{get_input_as_string(most_recent_input)}",
                        pos + len(get_input_as_string(previous_input)),
                    ),
                ],
            ),
        )
        self.most_recent_input = most_recent_input
        self.pos = pos
        self.previous_input = previous_input

    @property
    def src(self):
        return f"{get_input_as_string(self.previous_input)}{get_input_as_string(self.most_recent_input)}"

    @property
    def error_for_most_recent_input(self):
        return "\n".join(
            [
                INPUT_PARSER_ERROR_HEADER_MESSAGE,
                "",
                *build_error_position(
                    get_input_as_string(self.most_recent_input), self.pos,
                ),
            ],
        )



# import type { GraphNode, } from "./graph-node.js";
GraphNode = object


class RuleRef:
    __nodes__: set[GraphNode] | None = None
    value: int

    def __init__(self, value: int) -> None:
        self.value = value

    @property
    def nodes(self) -> set[GraphNode]:
        if self.__nodes__ is None:
            raise ValueError("Nodes are not set")
        return self.__nodes__

    @nodes.setter
    def nodes(self, nodes: set[GraphNode]) -> None:
        self.__nodes__ = nodes

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RuleRef) and self.value == other.value



from .colorize import Color, colorize


def test_colorize_strings():
    assert colorize("hello", Color.BLUE) == "\x1b[34mhello"
    assert colorize("test", Color.CYAN) == "\x1b[36mtest"
    assert colorize("example", Color.GREEN) == "\x1b[32mexample"


def test_colorize_numbers():
    assert colorize(123, Color.RED) == "\x1b[31m123"
    assert colorize(456, Color.GRAY) == "\x1b[90m456"
    assert colorize(789, Color.YELLOW) == "\x1b[33m789"



from typing import cast

import pytest

from ..grammar_graph.grammar_graph_types import (
    Range,
    RuleChar,
    RuleCharExclude,
    RuleEnd,
)
from ..grammar_graph.rule_ref import RuleRef
from ..rules_builder.rules_builder_types import (
    InternalRuleDef,
    InternalRuleDefAlt,
    InternalRuleDefChar,
    InternalRuleDefCharAlt,
    InternalRuleDefCharNot,
    InternalRuleDefCharRngUpper,
    InternalRuleDefEnd,
    InternalRuleDefReference,
)
from .build_rule_stack import build_rule_stack


def ichar(value: list[int]) -> InternalRuleDefChar:
    return InternalRuleDefChar(value=value)


def ichar_alt(value: int) -> InternalRuleDefCharAlt:
    return InternalRuleDefCharAlt(value=value)


def ichar_rng_upper(value: int) -> InternalRuleDefCharRngUpper:
    return InternalRuleDefCharRngUpper(value=value)


def ichar_not(value: list[int]) -> InternalRuleDefCharNot:
    return InternalRuleDefCharNot(value=value)


def ialt() -> InternalRuleDefAlt:
    return InternalRuleDefAlt()


def iend() -> InternalRuleDefEnd:
    return InternalRuleDefEnd()


def iref(value: int) -> InternalRuleDefReference:
    return InternalRuleDefReference(value=value)


def make_range(lower: int | str, upper: int | str) -> Range:
    return cast(
        Range,
        [
            lower if isinstance(lower, int) else ord(str(lower)),
            upper if isinstance(upper, int) else ord(str(upper)),
        ],
    )


def describe_build_rule_stack():
    def test_it_builds_rule_stack_for_a_single_path():
        assert build_rule_stack(
            [
                ichar([120]),
            ],
        ) == [
            [
                RuleChar(value=[120]),
                RuleEnd(),
            ],
        ]

    def test_it_builds_rule_stack_for_two_alternate_paths():
        result = build_rule_stack(
            [
                ichar([ord("x")]),
                ialt(),
                ichar([ord("y")]),
            ],
        )
        assert result == [
            [
                RuleChar(value=[ord("x")]),
                RuleEnd(),
            ],
            [
                RuleChar(value=[ord("y")]),
                RuleEnd(),
            ],
        ]

    def test_it_builds_rule_stack_for_three_alternate_paths():
        assert build_rule_stack(
            [
                ichar([ord("x")]),
                ialt(),
                ichar([ord("y")]),
                ialt(),
                ichar([ord("z")]),
            ],
        ) == [
            [
                RuleChar(value=[ord("x")]),
                RuleEnd(),
            ],
            [
                RuleChar(value=[ord("y")]),
                RuleEnd(),
            ],
            [
                RuleChar(value=[ord("z")]),
                RuleEnd(),
            ],
        ]

    def test_it_builds_rule_stack_for_char_not():
        assert build_rule_stack(
            [
                ichar_not([ord("x")]),
                ialt(),
                ichar_not([ord("y")]),
                ialt(),
                ichar_not([ord("z")]),
            ],
        ) == [
            [
                RuleCharExclude(value=[ord("x")]),
                RuleEnd(),
            ],
            [
                RuleCharExclude(value=[ord("y")]),
                RuleEnd(),
            ],
            [
                RuleCharExclude(value=[ord("z")]),
                RuleEnd(),
            ],
        ]

    def test_it_builds_rule_stack_for_mixed_char_and_char_not():
        assert build_rule_stack(
            [
                ichar_not([ord("x")]),
                ialt(),
                ichar([ord("y")]),
                ialt(),
                ichar_not([ord("z")]),
            ],
        ) == [
            [
                RuleCharExclude(value=[ord("x")]),
                RuleEnd(),
            ],
            [
                RuleChar(value=[ord("y")]),
                RuleEnd(),
            ],
            [
                RuleCharExclude(value=[ord("z")]),
                RuleEnd(),
            ],
        ]

    def test_it_builds_rule_stack_for_char_not_with_two_characters_and_a_range():
        result = build_rule_stack(
            [
                ichar_not([ord("x")]),
                ichar_alt(ord("y")),
                ichar_alt(ord("z")),
                ichar_rng_upper(130),
            ],
        )
        assert result == [
            [
                RuleCharExclude(value=[120, 121, make_range(122, 130)]),
                RuleEnd(),
            ],
        ]

    def describe_ranges():
        @pytest.mark.parametrize(
            ("grammar", "input", "expected"),
            [
                (
                    "[a-z]",
                    [
                        ichar([ord("a")]),
                        ichar_rng_upper(ord("z")),
                        iend(),
                    ],
                    [
                        [
                            RuleChar(value=[make_range("a", "z")]),
                            RuleEnd(),
                        ],
                    ],
                ),
                (
                    "[a-zA-Z]",
                    [
                        ichar([ord("a")]),
                        ichar_rng_upper(ord("z")),
                        ichar_alt(ord("A")),
                        ichar_rng_upper(ord("Z")),
                        iend(),
                    ],
                    [
                        [
                            RuleChar(
                                value=[
                                    make_range("a", "z"),
                                    make_range("A", "Z"),
                                ],
                            ),
                            RuleEnd(),
                        ],
                    ],
                ),
                (
                    "[a-zA-Z0-9]",
                    [
                        ichar([ord("a")]),
                        ichar_rng_upper(ord("z")),
                        ichar_alt(ord("A")),
                        ichar_rng_upper(ord("Z")),
                        ichar_alt(ord("0")),
                        ichar_rng_upper(ord("9")),
                        iend(),
                    ],
                    [
                        [
                            RuleChar(
                                value=[
                                    make_range("a", "z"),
                                    make_range("A", "Z"),
                                    make_range("0", "9"),
                                ],
                            ),
                            RuleEnd(),
                        ],
                    ],
                ),
            ],
        )
        def test_it_builds_rule_stack_for_a_char_with_no_modifiers(
            grammar,
            input,
            expected,
        ):
            result = build_rule_stack(input)
            assert result == expected

    @pytest.mark.parametrize(
        ("grammar", "input", "expected"),
        [
            (
                "[a-z]?",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ialt(),
                    iend(),
                ],
                [
                    [
                        RuleChar(value=[make_range("a", "z")]),
                        RuleEnd(),
                    ],
                    [
                        RuleEnd(),
                    ],
                ],
            ),
            (
                "[a-zA-Z]?",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    ialt(),
                    iend(),
                ],
                [
                    [
                        RuleChar(value=[make_range("a", "z"), make_range("A", "Z")]),
                        RuleEnd(),
                    ],
                    [
                        RuleEnd(),
                    ],
                ],
            ),
            (
                "[a-zA-Z0-9]?",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    ichar_alt(ord("0")),
                    ichar_rng_upper(ord("9")),
                    ialt(),
                    iend(),
                ],
                [
                    [
                        RuleChar(
                            value=[
                                make_range("a", "z"),
                                make_range("A", "Z"),
                                make_range("0", "9"),
                            ],
                        ),
                        RuleEnd(),
                    ],
                    [
                        RuleEnd(),
                    ],
                ],
            ),
        ],
    )
    def test_it_builds_rule_stack_for_a_char_with_question_mark_modifier(
        grammar,
        input,
        expected,
    ):
        assert build_rule_stack(input) == expected

    @pytest.mark.parametrize(
        ("grammar", "input", "expected"),
        [
            (
                "[a-z]+",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    iref(1),
                    ialt(),
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    iend(),
                ],
                [
                    [
                        RuleChar(value=[make_range("a", "z")]),
                        RuleRef(1),
                        RuleEnd(),
                    ],
                    [
                        RuleChar(value=[make_range("a", "z")]),
                        RuleEnd(),
                    ],
                ],
            ),
            (
                "[a-zA-Z]+",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    iref(1),
                    ialt(),
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    iend(),
                ],
                [
                    [
                        RuleChar(value=[make_range("a", "z"), make_range("A", "Z")]),
                        RuleRef(1),
                        RuleEnd(),
                    ],
                    [
                        RuleChar(value=[make_range("a", "z"), make_range("A", "Z")]),
                        RuleEnd(),
                    ],
                ],
            ),
            (
                "[a-zA-Z0-9]+",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    ichar_alt(ord("0")),
                    ichar_rng_upper(ord("9")),
                    iref(1),
                    ialt(),
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    ichar_alt(ord("0")),
                    ichar_rng_upper(ord("9")),
                    iend(),
                ],
                [
                    [
                        RuleChar(
                            value=[
                                make_range("a", "z"),
                                make_range("A", "Z"),
                                make_range("0", "9"),
                            ],
                        ),
                        RuleRef(1),
                        RuleEnd(),
                    ],
                    [
                        RuleChar(
                            value=[
                                make_range("a", "z"),
                                make_range("A", "Z"),
                                make_range("0", "9"),
                            ],
                        ),
                        RuleEnd(),
                    ],
                ],
            ),
        ],
    )
    def test_it_builds_rule_stack_for_a_char_with_plus_modifier(
        grammar,
        input,
        expected,
    ):
        result = build_rule_stack(input)
        assert result == expected

    @pytest.mark.parametrize(
        ("grammar", "input", "expected"),
        [
            (
                "[a-z]*",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    iref(1),
                    ialt(),
                    iend(),
                ],
                [
                    [
                        RuleChar(value=[make_range("a", "z")]),
                        RuleRef(1),
                        RuleEnd(),
                    ],
                    [
                        RuleEnd(),
                    ],
                ],
            ),
            (
                "[a-zA-Z]*",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    iref(1),
                    ialt(),
                    iend(),
                ],
                [
                    [
                        RuleChar(
                            value=[
                                make_range("a", "z"),
                                make_range("A", "Z"),
                            ],
                        ),
                        RuleRef(1),
                        RuleEnd(),
                    ],
                    [
                        RuleEnd(),
                    ],
                ],
            ),
            (
                "[a-zA-Z0-9]*",
                [
                    ichar([ord("a")]),
                    ichar_rng_upper(ord("z")),
                    ichar_alt(ord("A")),
                    ichar_rng_upper(ord("Z")),
                    ichar_alt(ord("0")),
                    ichar_rng_upper(ord("9")),
                    iref(1),
                    ialt(),
                    iend(),
                ],
                [
                    [
                        RuleChar(
                            value=[
                                make_range("a", "z"),
                                make_range("A", "Z"),
                                make_range("0", "9"),
                            ],
                        ),
                        RuleRef(1),
                        RuleEnd(),
                    ],
                    [
                        RuleEnd(),
                    ],
                ],
            ),
        ],
    )
    def test_it_builds_rule_stack_for_a_char_with_asterisk_modifier(
        grammar,
        input,
        expected,
    ):
        result = build_rule_stack(input)
        assert result == expected

    def test_char_with_range():
        input: list[InternalRuleDef] = [
            ichar([ord("a")]),
            ichar_rng_upper(ord("z")),
            ichar_alt(ord("A")),
            ichar_rng_upper(ord("Z")),
            ichar_alt(ord("_")),
            iend(),
        ]

        assert build_rule_stack(input) == [
            [
                RuleChar(
                    value=[
                        make_range("a", "z"),
                        make_range("A", "Z"),
                        ord("_"),
                    ],
                ),
                RuleEnd(),
            ],
        ]

    def test_it_builds_rule_stack_for_situation_root_ws_plus_newline_ws_star():
        input: list[InternalRuleDef] = [
            ichar([32]),
            ichar_alt(92),
            ichar_alt(110),
            iref(4),
            ialt(),
            iend(),
        ]

        expected = [
            [
                RuleChar(value=[32, 92, 110]),
                RuleRef(4),
                RuleEnd(),
            ],
            [
                RuleEnd(),
            ],
        ]

        assert build_rule_stack(input) == expected

    def test_it_builds_rule_stack_for_situation_root_char_plus_char_range():
        input: list[InternalRuleDef] = [
            ichar([ord("a")]),
            ichar_rng_upper(ord("z")),
            ichar_alt(ord("A")),
            ichar_rng_upper(ord("Z")),
            iref(1),
            ialt(),
            ichar([ord("a")]),
            ichar_rng_upper(ord("z")),
            ichar_alt(ord("A")),
            ichar_rng_upper(ord("Z")),
            iend(),
        ]

        expected = [
            [
                RuleChar(
                    value=[
                        make_range("a", "z"),
                        make_range("A", "Z"),
                    ],
                ),
                RuleRef(1),
                RuleEnd(),
            ],
            [
                RuleChar(
                    value=[
                        make_range("a", "z"),
                        make_range("A", "Z"),
                    ],
                ),
                RuleEnd(),
            ],
        ]

        assert build_rule_stack(input) == expected



from time import perf_counter

from ..utils.errors import GrammarParseError
from .is_word_char import is_word_char
from .parse_char import parse_char
from .parse_name import parse_name
from .parse_space import parse_space
from .rules_builder_types import (
    InternalRuleDef,
    InternalRuleDefAlt,
    InternalRuleDefChar,
    InternalRuleDefCharAlt,
    InternalRuleDefCharNot,
    InternalRuleDefCharRngUpper,
    InternalRuleDefEnd,
    InternalRuleDefReference,
)
from .symbol_ids import SymbolIds


def get_out_elements(
    type_of_rule: type[InternalRuleDef],
    startchar_value: int,
) -> InternalRuleDef:
    if type_of_rule == InternalRuleDefChar:
        return InternalRuleDefChar(value=[startchar_value])
    if type_of_rule == InternalRuleDefCharNot:
        return InternalRuleDefCharNot(value=[startchar_value])
    if type_of_rule == InternalRuleDefCharRngUpper:
        return InternalRuleDefCharRngUpper(value=startchar_value)
    if type_of_rule == InternalRuleDefAlt:
        return InternalRuleDefAlt()
    if type_of_rule == InternalRuleDefEnd:
        return InternalRuleDefEnd()
    if type_of_rule == InternalRuleDefCharAlt:
        return InternalRuleDefCharAlt(value=startchar_value)

    raise ValueError(f"Invalid type: {type_of_rule}")


class RulesBuilder:
    pos: int
    symbol_ids: SymbolIds
    rules: list[list[InternalRuleDef]]
    src: str
    start: float
    time_limit: int

    def __init__(self, src: str, limit: int = 1000):
        self.pos = 0
        self.symbol_ids = SymbolIds()
        self.rules = []
        self.src = src
        self.start = perf_counter()
        self.time_limit = limit
        self.parse(src)

    def parse(self, src: str) -> None:
        self.pos = parse_space(src, 0, True)
        while self.pos < len(src):
            self.parse_rule(src)

        # Validate the state to ensure that all rules are defined
        for rule in self.rules:
            for elem in rule:
                if isinstance(elem, InternalRuleDefReference):
                    rule_exists = (
                        elem.value < len(self.rules) and len(self.rules[elem.value]) > 0
                    )
                    if not rule_exists:
                        missing_rule_name = self.symbol_ids.reverse_get(elem.value)
                        missing_rule_pos = self.symbol_ids.get_pos(missing_rule_name)

                        # Skip over the ::= and any whitespace
                        while missing_rule_pos < len(src) and (
                            src[missing_rule_pos] == ":"
                            or src[missing_rule_pos] == "="
                            or src[missing_rule_pos].isspace()
                        ):
                            missing_rule_pos += 1

                        raise GrammarParseError(
                            src,
                            missing_rule_pos,
                            f'Undefined rule identifier "{missing_rule_name}"',
                        )

    def parse_rule(self, src: str) -> None:
        name = parse_name(src, self.pos)
        self.pos = parse_space(src, self.pos + len(name), False)
        rule_id = self.get_symbol_id(name, len(name))

        self.pos = parse_space(src, self.pos, True)
        if not (
            self.pos + 2 < len(src)  # Ensure the position + 2 is within bounds
            and src[self.pos] == ":"
            and src[self.pos + 1] == ":"
            and src[self.pos + 2] == "="
        ):
            raise GrammarParseError(src, self.pos, f"Expecting ::= at {self.pos}")
        self.pos += 3
        self.pos = parse_space(src, self.pos, True)

        self.parse_alternates(name, rule_id)

        # Check if self.pos is within the bounds of src before checking for a carriage return
        if self.pos < len(src) and src[self.pos] == "\r":
            self.pos += 2 if src[self.pos + 1] == "\n" else 1
        elif self.pos < len(src) and src[self.pos] == "\n":
            self.pos += 1
        elif self.pos < len(src) and src[self.pos]:
            raise GrammarParseError(
                src,
                self.pos,
                f"Expecting newline or end at {self.pos}",
            )
        self.pos = parse_space(src, self.pos, True)

    def get_symbol_id(self, src: str, length: int) -> int:
        next_id = len(self.symbol_ids)
        key = src[:length]
        if key not in self.symbol_ids:
            self.symbol_ids.set(key, next_id, self.pos)
        return self.symbol_ids[key]

    def generate_symbol_id(self, base_name: str) -> int:
        next_id = len(self.symbol_ids)
        self.symbol_ids.set(f"{base_name}_{next_id}", next_id, self.pos)
        return next_id

    def add_rule(self, rule_id: int, rule: list[InternalRuleDef]) -> None:
        while len(self.rules) <= rule_id:
            self.rules.append([])
        self.rules[rule_id] = rule

    def check_duration(self) -> None:
        if perf_counter() - self.start > self.time_limit:
            raise GrammarParseError(
                self.src,
                self.pos,
                f"Duration of {self.time_limit} exceeded",
            )

    def parse_sequence(
        self,
        rule_name: str,
        out_elements: list[InternalRuleDef],
        depth: int = 0,
    ) -> None:
        is_nested = depth != 0
        src = self.src
        last_sym_start = len(out_elements)

        while self.pos < len(src):
            if src[self.pos] == '"':
                self.pos += 1
                last_sym_start = len(out_elements)
                while src[self.pos] != '"':
                    self.check_duration()
                    value, inc_pos = parse_char(src, self.pos)
                    out_elements.append(
                        InternalRuleDefChar(
                            value=[value],
                        ),
                    )
                    self.pos += inc_pos
                self.pos = parse_space(src, self.pos + 1, is_nested)
            elif src[self.pos] == "[":
                self.pos += 1
                start_type: type[InternalRuleDef] = InternalRuleDefChar
                if src[self.pos] == "^":
                    self.pos += 1
                    start_type = InternalRuleDefCharNot
                last_sym_start = len(out_elements)
                while src[self.pos] != "]":
                    self.check_duration()
                    type_: type[InternalRuleDef] = (
                        InternalRuleDefCharAlt
                        if last_sym_start < len(out_elements)
                        else start_type
                    )
                    startchar_value, inc_pos = parse_char(src, self.pos)
                    self.pos += inc_pos
                    out_elements.append(get_out_elements(type_, startchar_value))

                    if src[self.pos] == "-" and src[self.pos + 1] != "]":
                        self.pos += 1
                        endchar_value, inc_pos = parse_char(src, self.pos)
                        out_elements.append(
                            InternalRuleDefCharRngUpper(
                                value=endchar_value,
                            ),
                        )
                        self.pos += inc_pos
                self.pos = parse_space(src, self.pos + 1, is_nested)
            elif is_word_char(src[self.pos]):
                name = parse_name(src, self.pos)
                ref_rule_id = self.get_symbol_id(name, len(name))
                self.pos += len(name)
                self.pos = parse_space(src, self.pos, is_nested)

                last_sym_start = len(out_elements)
                out_elements.append(
                    InternalRuleDefReference(
                        value=ref_rule_id,
                    ),
                )
            elif src[self.pos] == "(":
                self.pos = parse_space(src, self.pos + 1, True)
                sub_rule_id = self.generate_symbol_id(rule_name)
                self.parse_alternates(rule_name, sub_rule_id, depth + 1)
                last_sym_start = len(out_elements)
                out_elements.append(
                    InternalRuleDefReference(
                        value=sub_rule_id,
                    ),
                )
                if src[self.pos] != ")":
                    raise GrammarParseError(
                        src,
                        self.pos,
                        f"Expecting ')' at {self.pos}",
                    )
                self.pos = parse_space(src, self.pos + 1, is_nested)
            elif src[self.pos] in "*+?":
                if last_sym_start == len(out_elements):
                    raise GrammarParseError(
                        src,
                        self.pos,
                        f"Expecting preceding item to */+/? at {self.pos}",
                    )
                sub_rule_id = self.generate_symbol_id(rule_name)
                sub_rule = out_elements[last_sym_start:]
                if src[self.pos] in "*+":
                    sub_rule.append(
                        InternalRuleDefReference(
                            value=sub_rule_id,
                        ),
                    )
                sub_rule.append(InternalRuleDefAlt())
                if src[self.pos] == "+":
                    sub_rule.extend(out_elements[last_sym_start:])
                sub_rule.append(InternalRuleDefEnd())
                self.add_rule(sub_rule_id, sub_rule)
                out_elements[last_sym_start:] = [
                    InternalRuleDefReference(
                        value=sub_rule_id,
                    ),
                ]
                self.pos = parse_space(src, self.pos + 1, is_nested)
            else:
                break

    def parse_alternates(self, rule_name: str, rule_id: int, depth: int = 0) -> None:
        src = self.src
        rule = []
        self.parse_sequence(rule_name, rule, depth)
        # Ensure that self.pos is within bounds before checking src[self.pos]
        while self.pos < len(src) and src[self.pos] == "|":
            self.check_duration()
            rule.append(InternalRuleDefAlt())
            self.pos = parse_space(src, self.pos + 1, True)
            self.parse_sequence(rule_name, rule, depth)
        rule.append(InternalRuleDefEnd())
        self.add_rule(rule_id, rule)



import pytest

from .rules_builder import RulesBuilder
from .rules_builder_types import (
    InternalRuleDefAlt,
    InternalRuleDefChar,
    InternalRuleDefCharAlt,
    InternalRuleDefCharNot,
    InternalRuleDefCharRngUpper,
    InternalRuleDefEnd,
    InternalRuleDefReference,
)

test_cases = [
    (
        "single-string",
        'root ::= "foo"',
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefChar(value=[ord("f")]),
                    InternalRuleDefChar(value=[ord("o")]),
                    InternalRuleDefChar(value=[ord("o")]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "quote character",
        r'root ::= "\""',
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefChar(value=[ord('"')]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "two-lines-referencing-expression",
        '''root ::= foo
            foo ::= "bar"''',
        (
            [("root", 0), ("foo", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("b")]),
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefChar(value=[ord("r")]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "expression with dash",
        '''root ::= foo-bar
            foo-bar ::= "bar"''',
        (
            [("root", 0), ("foo-bar", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("b")]),
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefChar(value=[ord("r")]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "simple-grammar",
        """
            root  ::= (expr "=" term "\n")+
            expr  ::= term ([-+*/] term)*
            term  ::= [0-9]+
            """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("term", 3),
                ("root_4", 4),
                ("expr_5", 5),
                ("expr_6", 6),
                ("term_7", 7),
            ],
            [
                [
                    InternalRuleDefReference(value=4),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=2),
                    InternalRuleDefChar(value=[61]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefChar(value=[10]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=3),
                    InternalRuleDefReference(value=6),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[45]),
                    InternalRuleDefCharAlt(value=43),
                    InternalRuleDefCharAlt(value=42),
                    InternalRuleDefCharAlt(value=47),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=5),
                    InternalRuleDefReference(value=6),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "longer-grammar",
        """
            root  ::= (expr "=" ws term "\n")+
            expr  ::= term ([-+*/] term)*
            term  ::= ident | num | "(" ws expr ")" ws
            ident ::= [a-z] [a-z0-9_]* ws
            num   ::= [0-9]+ ws
            ws    ::= [ \t\n]*
            """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("ws", 3),
                ("term", 4),
                ("root_5", 5),
                ("expr_6", 6),
                ("expr_7", 7),
                ("ident", 8),
                ("num", 9),
                ("ident_10", 10),
                ("num_11", 11),
                ("ws_12", 12),
            ],
            [
                [
                    InternalRuleDefReference(value=5),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=2),
                    InternalRuleDefChar(value=[61]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefChar(value=[10]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=4),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=12),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=8),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=9),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[40]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefChar(value=[41]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=5),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[45]),
                    InternalRuleDefCharAlt(value=43),
                    InternalRuleDefCharAlt(value=42),
                    InternalRuleDefCharAlt(value=47),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=6),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefReference(value=10),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=11),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefCharAlt(value=48),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=95),
                    InternalRuleDefReference(value=10),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=11),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[32]),
                    InternalRuleDefCharAlt(value=9),
                    InternalRuleDefCharAlt(value=10),
                    InternalRuleDefReference(value=12),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "character unicode grammar",
        "root  ::= [ぁ-ゟ]",
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefChar(value=[ord("ぁ")]),
                    InternalRuleDefCharRngUpper(value=ord("ゟ")),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "character alts grammar",
        "root  ::= [az]",
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefCharAlt(value=ord("z")),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "character range grammar",
        "root  ::= [a-zA-Z0-9]",
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefCharRngUpper(value=ord("z")),
                    InternalRuleDefCharAlt(value=ord("A")),
                    InternalRuleDefCharRngUpper(value=ord("Z")),
                    InternalRuleDefCharAlt(value=ord("0")),
                    InternalRuleDefCharRngUpper(value=ord("9")),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "character range grammar with dash at end",
        "root  ::= [a-z-]",
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefCharRngUpper(value=ord("z")),
                    InternalRuleDefCharAlt(value=ord("-")),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "grouping",
        'root  ::= "f" ("b" | "a")',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefChar(value=[ord("f")]),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("b")]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "optional",
        'root  ::= "f"?',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("f")]),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "repeating",
        'root  ::= "f"*',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("f")]),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "repeating-at-least-once",
        'root  ::= "f"+',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("f")]),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[ord("f")]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "character range with optional repeating",
        "root  ::= [a-zA-Z0-9]*",
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefCharRngUpper(value=ord("z")),
                    InternalRuleDefCharAlt(value=ord("A")),
                    InternalRuleDefCharRngUpper(value=ord("Z")),
                    InternalRuleDefCharAlt(value=ord("0")),
                    InternalRuleDefCharRngUpper(value=ord("9")),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "grouping-repeating",
        'root  ::= "f" ("b" | "a")*',
        (
            [("root", 0), ("root_1", 1), ("root_2", 2)],
            [
                [
                    InternalRuleDefChar(value=[ord("f")]),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[ord("b")]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[ord("a")]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "negation",
        r"root ::= [^\n]",
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefCharNot(value=[ord("\n")]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "negation of range",
        "root ::= [^0-9]",
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefCharNot(value=[ord("0")]),
                    InternalRuleDefCharRngUpper(value=ord("9")),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "negation with after",
        r'root ::= [^\n]+ "\n"',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefChar(value=[ord("\n")]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefCharNot(value=[ord("\n")]),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefAlt(),
                    InternalRuleDefCharNot(value=[ord("\n")]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "longer negation",
        r'root ::= "\"" ( [^"abcdefgh])* ',
        (
            [("root", 0), ("root_1", 1), ("root_2", 2)],
            [
                [
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefCharNot(value=[34]),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharAlt(value=98),
                    InternalRuleDefCharAlt(value=99),
                    InternalRuleDefCharAlt(value=100),
                    InternalRuleDefCharAlt(value=101),
                    InternalRuleDefCharAlt(value=102),
                    InternalRuleDefCharAlt(value=103),
                    InternalRuleDefCharAlt(value=104),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "longer negation with a range",
        r'root ::= "\"" ( [^"abcdefghA-Z])* ',
        (
            [("root", 0), ("root_1", 1), ("root_2", 2)],
            [
                [
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefCharNot(value=[34]),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharAlt(value=98),
                    InternalRuleDefCharAlt(value=99),
                    InternalRuleDefCharAlt(value=100),
                    InternalRuleDefCharAlt(value=101),
                    InternalRuleDefCharAlt(value=102),
                    InternalRuleDefCharAlt(value=103),
                    InternalRuleDefCharAlt(value=104),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=ord("Z")),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    *[
        (
            key,
            f'root ::= "{escaped_char}"',
            (
                [("root", 0)],
                [
                    [
                        InternalRuleDefChar(value=[actual_char]),
                        InternalRuleDefEnd(),
                    ],
                ],
            ),
        )
        for key, escaped_char, actual_char in [
            ("escaped 8-bit unicode char", r"\x2A", ord("\x2A")),
            ("escaped 16-bit unicode char", r"\u006F", ord("\u006F")),
            ("escaped 32-bit unicode char", r"\U0001F4A9", 128169),
            ("escaped tab char", r"\t", ord("\t")),
            ("escaped new line char", r"\n", ord("\n")),
            ("escaped \r char", r"\r", ord("\r")),
            ("escaped quote char", r"\"", ord('"')),
            ("escaped [ char", r"\[", ord("[")),
            ("escaped ] char", r"\]", ord("]")),
            ("escaped \\ char", r"\\", ord("\\")),
        ]
    ],
    (
        "simple arithmetic",
        r"""
            root ::= (expr "=" term "\n")+
            expr ::= term ([-+*/] term)*
            term ::= [0-9]+
            """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("term", 3),
                ("root_4", 4),
                ("expr_5", 5),
                ("expr_6", 6),
                ("term_7", 7),
            ],
            [
                [
                    InternalRuleDefReference(value=4),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=2),
                    InternalRuleDefChar(value=[61]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefChar(value=[10]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=3),
                    InternalRuleDefReference(value=6),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[45]),
                    InternalRuleDefCharAlt(value=43),
                    InternalRuleDefCharAlt(value=42),
                    InternalRuleDefCharAlt(value=47),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=5),
                    InternalRuleDefReference(value=6),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "ranges with chars",
        """
            root ::= [a-z0-9_]*
            """,
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefCharAlt(value=48),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=95),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "nested ranges with chars",
        """
            root ::= [a-z] [a-z0-9_]*
            """,
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefCharAlt(value=48),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=95),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "expression with nested range with chars",
        r"""
            root ::= ident
            ident ::= [a-z] [a-z0-9_]* ws
            ws ::= [ \t\n]*
            """,
        (
            [("root", 0), ("ident", 1), ("ident_2", 2), ("ws", 3), ("ws_4", 4)],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefCharAlt(value=48),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=95),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=4),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[32]),
                    InternalRuleDefCharAlt(value=9),
                    InternalRuleDefCharAlt(value=10),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "lots of escapes",
        r'root ::= "\x2A" "\u006F" "\U0001F4A9" "\t" "\n" "\r" "\"" "\[" "\]" "\\"',
        (
            [("root", 0)],
            [
                [
                    InternalRuleDefChar(value=[42]),  # \x2A
                    InternalRuleDefChar(value=[111]),  # \u006F
                    InternalRuleDefChar(value=[128169]),  # \U0001F4A9
                    InternalRuleDefChar(value=[9]),  # \t
                    InternalRuleDefChar(value=[10]),  # \n
                    InternalRuleDefChar(value=[13]),  # \r
                    InternalRuleDefChar(value=[34]),  # \"
                    InternalRuleDefChar(value=[91]),  # \[
                    InternalRuleDefChar(value=[93]),  # \]
                    InternalRuleDefChar(value=[92]),  # \\
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "lots of escape and alternate escapes",
        r"""root ::= "\x2A" "\u006F" "\U0001F4A9" "\t" "\n" "\r" "\"" "\[" "\]" "\\" (
                "\x2A" | "\u006F" | "\U0001F4A9" | "\t" | "\n" | "\r" | "\"" | "\[" | "\]"  | "\\" )""",
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    InternalRuleDefChar(value=[42]),
                    InternalRuleDefChar(value=[111]),
                    InternalRuleDefChar(value=[128169]),
                    InternalRuleDefChar(value=[9]),
                    InternalRuleDefChar(value=[10]),
                    InternalRuleDefChar(value=[13]),
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefChar(value=[91]),
                    InternalRuleDefChar(value=[93]),
                    InternalRuleDefChar(value=[92]),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[42]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[111]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[128169]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[9]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[10]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[13]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[91]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[93]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[92]),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "arithmetic",
        r"""
            root  ::= (expr "=" ws term "\n")+
            expr  ::= term ([-+*/] term)*
            term  ::= ident | num | "(" ws expr ")" ws
            ident ::= [a-z] [a-z0-9_]* ws
            num   ::= [0-9]+ ws
            ws    ::= [ \t\n]*
            """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("ws", 3),
                ("term", 4),
                ("root_5", 5),
                ("expr_6", 6),
                ("expr_7", 7),
                ("ident", 8),
                ("num", 9),
                ("ident_10", 10),
                ("num_11", 11),
                ("ws_12", 12),
            ],
            [
                [
                    InternalRuleDefReference(value=5),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=2),
                    InternalRuleDefChar(value=[61]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefChar(value=[10]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=4),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=12),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=8),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=9),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[40]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefChar(value=[41]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=5),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[45]),
                    InternalRuleDefCharAlt(value=43),
                    InternalRuleDefCharAlt(value=42),
                    InternalRuleDefCharAlt(value=47),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=6),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefReference(value=10),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=11),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefCharRngUpper(value=122),
                    InternalRuleDefCharAlt(value=48),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=95),
                    InternalRuleDefReference(value=10),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=11),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[32]),
                    InternalRuleDefCharAlt(value=9),
                    InternalRuleDefCharAlt(value=10),
                    InternalRuleDefReference(value=12),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "json.gbnf (string)",
        r"""
            root ::=
            "\"" (
                [^"\\\x7F\x00-\x1F] |
                "\\" (["\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F]) # escapes
            )* "\""
            """,
        (
            [("root", 0), ("root_1", 1), ("root_2", 2), ("root_3", 3)],
            [
                [
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefCharNot(value=[34]),
                    InternalRuleDefCharAlt(value=92),
                    InternalRuleDefCharAlt(value=127),
                    InternalRuleDefCharAlt(value=0),
                    InternalRuleDefCharRngUpper(value=31),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[92]),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefCharAlt(value=92),
                    InternalRuleDefCharAlt(value=47),
                    InternalRuleDefCharAlt(value=98),
                    InternalRuleDefCharAlt(value=102),
                    InternalRuleDefCharAlt(value=110),
                    InternalRuleDefCharAlt(value=114),
                    InternalRuleDefCharAlt(value=116),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[117]),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "json.gbnf (full)",
        r"""
            root   ::= object
            value  ::= object | array | string | number | ("true" | "false" | "null") ws
            object ::=
              "{" ws (
                        string ":" ws value
                ("," ws string ":" ws value)*
              )? "}" ws
            array  ::=
              "[" ws (
                        value
                ("," ws value)*
              )? "]" ws
                  string ::=
              "\"" (
                [^"\\\x7F\x00-\x1F] |
                "\\" (["\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F]) # escapes
              )* "\"" ws
            number ::= ("-"? ([0-9] | [1-9] [0-9]*)) ("." [0-9]+)? ([eE] [-+]? [0-9]+)? ws
            # Optional space: by convention, applied in this grammar after literal chars when allowed
            ws ::= ([ \t\n] ws)?
            """,
        (
            [
                ("root", 0),
                ("object", 1),
                ("value", 2),
                ("array", 3),
                ("string", 4),
                ("number", 5),
                ("value_6", 6),
                ("ws", 7),
                ("object_8", 8),
                ("object_9", 9),
                ("object_10", 10),
                ("object_11", 11),
                ("array_12", 12),
                ("array_13", 13),
                ("array_14", 14),
                ("array_15", 15),
                ("string_16", 16),
                ("string_17", 17),
                ("string_18", 18),
                ("number_19", 19),
                ("number_20", 20),
                ("number_21", 21),
                ("number_22", 22),
                ("number_23", 23),
                ("number_24", 24),
                ("number_25", 25),
                ("number_26", 26),
                ("number_27", 27),
                ("number_28", 28),
                ("number_29", 29),
                ("ws_30", 30),
                ("ws_31", 31),
            ],
            [
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[123]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefReference(value=11),
                    InternalRuleDefChar(value=[125]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=3),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=5),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=6),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[91]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefReference(value=15),
                    InternalRuleDefChar(value=[93]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefReference(value=18),
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=19),
                    InternalRuleDefReference(value=25),
                    InternalRuleDefReference(value=29),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[116]),
                    InternalRuleDefChar(value=[114]),
                    InternalRuleDefChar(value=[117]),
                    InternalRuleDefChar(value=[101]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[102]),
                    InternalRuleDefChar(value=[97]),
                    InternalRuleDefChar(value=[108]),
                    InternalRuleDefChar(value=[115]),
                    InternalRuleDefChar(value=[101]),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[110]),
                    InternalRuleDefChar(value=[117]),
                    InternalRuleDefChar(value=[108]),
                    InternalRuleDefChar(value=[108]),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=31),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=4),
                    InternalRuleDefChar(value=[58]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefReference(value=10),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[44]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefChar(value=[58]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=9),
                    InternalRuleDefReference(value=10),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=8),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=2),
                    InternalRuleDefReference(value=14),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[44]),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=13),
                    InternalRuleDefReference(value=14),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=12),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefCharNot(value=[34]),
                    InternalRuleDefCharAlt(value=92),
                    InternalRuleDefCharAlt(value=127),
                    InternalRuleDefCharAlt(value=0),
                    InternalRuleDefCharRngUpper(value=31),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[92]),
                    InternalRuleDefReference(value=17),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[34]),
                    InternalRuleDefCharAlt(value=92),
                    InternalRuleDefCharAlt(value=47),
                    InternalRuleDefCharAlt(value=98),
                    InternalRuleDefCharAlt(value=102),
                    InternalRuleDefCharAlt(value=110),
                    InternalRuleDefCharAlt(value=114),
                    InternalRuleDefCharAlt(value=116),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[117]),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefCharAlt(value=97),
                    InternalRuleDefCharRngUpper(value=102),
                    InternalRuleDefCharAlt(value=65),
                    InternalRuleDefCharRngUpper(value=70),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=16),
                    InternalRuleDefReference(value=18),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=20),
                    InternalRuleDefReference(value=21),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[45]),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[49]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=22),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=22),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[46]),
                    InternalRuleDefReference(value=24),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=24),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=23),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[101]),
                    InternalRuleDefCharAlt(value=69),
                    InternalRuleDefReference(value=27),
                    InternalRuleDefReference(value=28),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[45]),
                    InternalRuleDefCharAlt(value=43),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefReference(value=28),
                    InternalRuleDefAlt(),
                    InternalRuleDefChar(value=[48]),
                    InternalRuleDefCharRngUpper(value=57),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=26),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[32]),
                    InternalRuleDefCharAlt(value=9),
                    InternalRuleDefCharAlt(value=10),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=30),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
    (
        "japanese",
        r"""
            # A probably incorrect grammar for Japanese
            root        ::= jp-char+ ([ \t\n] jp-char+)*
            jp-char     ::= hiragana | katakana | punctuation | cjk
            hiragana    ::= [ぁ-ゟ]
            katakana    ::= [ァ-ヿ]
            punctuation ::= [、-〾]
            cjk         ::= [一-鿿]
            """,
        (
            [
                ("root", 0),
                ("jp-char", 1),
                ("root_2", 2),
                ("root_3", 3),
                ("root_4", 4),
                ("root_5", 5),
                ("hiragana", 6),
                ("katakana", 7),
                ("punctuation", 8),
                ("cjk", 9),
            ],
            [
                [
                    InternalRuleDefReference(value=2),
                    InternalRuleDefReference(value=5),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=6),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=7),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=8),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=9),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=2),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[32]),
                    InternalRuleDefCharAlt(value=9),
                    InternalRuleDefCharAlt(value=10),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=1),
                    InternalRuleDefReference(value=4),
                    InternalRuleDefAlt(),
                    InternalRuleDefReference(value=1),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefReference(value=3),
                    InternalRuleDefReference(value=5),
                    InternalRuleDefAlt(),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[12353]),
                    InternalRuleDefCharRngUpper(value=12447),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[12449]),
                    InternalRuleDefCharRngUpper(value=12543),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[12289]),
                    InternalRuleDefCharRngUpper(value=12350),
                    InternalRuleDefEnd(),
                ],
                [
                    InternalRuleDefChar(value=[19968]),
                    InternalRuleDefCharRngUpper(value=40959),
                    InternalRuleDefEnd(),
                ],
            ],
        ),
    ),
]


@pytest.mark.parametrize(("key", "grammar", "expected"), test_cases)
def test_grammar_parser(key, grammar, expected):
    symbol_ids_expected, rules_expected = expected
    parsed_grammar = RulesBuilder(grammar.replace("\\n", "\n"))
    assert parsed_grammar.rules == rules_expected
    assert list(parsed_grammar.symbol_ids.items()) == symbol_ids_expected



import json
from typing import cast

from ..grammar_graph.grammar_graph_types import (
    Range,
    RuleChar,
    RuleCharExclude,
    RuleEnd,
    UnresolvedRule,
)
from ..grammar_graph.rule_ref import RuleRef
from ..grammar_graph.type_guards import is_range, is_rule_end
from ..rules_builder.rules_builder_types import (
    InternalRuleDef,
    InternalRuleDefChar,
    InternalRuleDefCharAlt,
    InternalRuleDefCharNot,
    InternalRuleDefCharRngUpper,
    InternalRuleDefReference,
    is_rule_def_alt,
    is_rule_def_char,
    is_rule_def_char_alt,
    is_rule_def_char_not,
    is_rule_def_char_rng_upper,
    is_rule_def_end,
    is_rule_def_ref,
)


def make_char_rule(
    rule_def: InternalRuleDefChar | InternalRuleDefCharNot,
) -> RuleChar | RuleCharExclude:
    value = cast(list[int | Range], rule_def.value)
    if is_rule_def_char_not(rule_def):
        return RuleCharExclude(value=value)
    if is_rule_def_char(rule_def):
        return RuleChar(value=value)

    raise ValueError(f"Unsupported rule type for make_char_rule: {rule_def}")


def build_rule_stack(linear_rules: list[InternalRuleDef]) -> list[list[UnresolvedRule]]:
    paths: list[UnresolvedRule] = []
    stack: list[list[UnresolvedRule]] = []
    idx = 0

    while idx < len(linear_rules):
        rule_def = linear_rules[idx]
        if is_rule_def_char(rule_def) or is_rule_def_char_not(rule_def):
            # this could be a single char, or a range, or a sequence of alts; we don't know until we step through it.
            rule_def = (
                cast(InternalRuleDefChar, rule_def)
                if is_rule_def_char(rule_def)
                else cast(InternalRuleDefCharNot, rule_def)
            )
            char_rule = make_char_rule(rule_def)
            idx += 1
            rule = linear_rules[idx] if idx < len(linear_rules) else None
            while idx < len(linear_rules) and (
                is_rule_def_char_rng_upper(rule) or is_rule_def_char_alt(rule)
            ):
                if is_rule_def_char_rng_upper(rule):
                    # previous rule value should be a number
                    prev_value = char_rule.value.pop()
                    if is_range(prev_value):
                        raise ValueError(
                            f"Unexpected range, expected a number but got an array: {json.dumps(prev_value)}",
                        )
                    if prev_value is None:
                        raise ValueError("Unexpected undefined value")

                    char_rule.value.append(
                        cast(
                            Range,
                            [prev_value, cast(InternalRuleDefCharRngUpper, rule).value],
                        ),
                    )
                if is_rule_def_char_alt(rule):
                    rule = cast(InternalRuleDefCharAlt, rule)
                    char_rule.value.append(cast(int, rule.value))
                idx += 1
                rule = linear_rules[idx] if idx < len(linear_rules) else None
            paths.append(char_rule)
        else:
            if is_rule_def_alt(rule_def):
                if len(paths) == 0:
                    raise ValueError("Encountered alt without anything before it")
                paths.append(RuleEnd())
                stack.append(paths)
                paths = []
            elif is_rule_def_end(rule_def):
                paths.append(RuleEnd())
            elif is_rule_def_ref(rule_def):
                paths.append(
                    RuleRef(cast(int, cast(InternalRuleDefReference, rule_def).value)),
                )
            elif is_rule_def_char_alt(rule_def):
                raise ValueError(
                    f"Encountered char alt, should be handled by above block: {rule_def}",
                )
            else:
                raise ValueError(f"Unsupported rule type: {rule_def}")
            idx += 1

    if not is_rule_end(paths[-1]):
        paths.append(RuleEnd())

    stack.append(paths)
    return stack



from .grammar_graph.graph import Graph
from .grammar_graph.parse_state import ParseState
from .grammar_parser.build_rule_stack import build_rule_stack
from .rules_builder import GrammarParseError, RulesBuilder

# ValidInput = str | int | float | list[int | float]


def GBNF(grammar: str, initial_string: str = ""):
    #   const grammar = typeof input === 'string' ? input : input.toString();

    rules_builder = RulesBuilder(grammar)
    rules, symbol_ids = rules_builder.rules, rules_builder.symbol_ids
    if len(rules) == 0:
        raise GrammarParseError(grammar, 0, "No rules were found")
    if symbol_ids["root"] is None:
        raise GrammarParseError(grammar, 0, "Grammar does not contain a 'root' symbol")
    root_id: int = symbol_ids["root"]

    stacked_rules = [build_rule_stack(rule) for rule in rules]
    graph = Graph(grammar, stacked_rules, root_id)
    return ParseState(graph, graph.add(initial_string))



class SymbolIds:
    __map__: dict[str, int]
    __pos__: dict[str, int]
    __reverse_map__: dict[int, str]

    def __init__(self):
        self.__map__ = {}
        self.__pos__ = {}
        self.__reverse_map__ = {}

    def items(self):
        return self.__map__.items()

    def __iter__(self):
        return iter(self.__map__.items())

    def __len__(self):
        return len(self.__map__)

    def __getitem__(self, key):
        return self.__map__[key]

    def set(self, key: str, value: int, pos: int):
        self.__map__[key] = value
        self.__reverse_map__[value] = key
        self.__pos__[key] = pos

    def __contains__(self, key: str):
        return key in self.__map__

    def reverse_get(self, key: int):
        val = self.__reverse_map__.get(key)
        if val is None:
            raise ValueError(f"SymbolIds does not contain value: {key}")
        return val

    def get_pos(self, key: str):
        val = self.__pos__.get(key)
        if val is None:
            raise ValueError(f"SymbolIds does not contain key: {key}")
        return val



def parse_space(src: str, pos: int, newline_ok: bool) -> int:
    while pos < len(src) and (
        src[pos] == " "
        or src[pos] == "\t"
        or src[pos] == "#"
        or (newline_ok and (src[pos] == "\r" or src[pos] == "\n"))
    ):
        if src[pos] == "#":
            while pos < len(src) and src[pos] not in ("\r", "\n"):
                pos += 1
        else:
            pos += 1
    return pos



from .build_error_position import build_error_position


def GRAMMAR_PARSER_ERROR_HEADER_MESSAGE(reason: str) -> str:
    return f"Failed to parse grammar: {reason}"


class GrammarParseError(Exception):
    grammar: str
    pos: int
    reason: str

    def __init__(self, grammar: str, pos: int, reason: str):
        super().__init__(
            "\n".join(
                [
                    GRAMMAR_PARSER_ERROR_HEADER_MESSAGE(reason),
                    "",
                    *build_error_position(grammar, pos),
                ],
            ),
        )
        self.grammar = grammar
        self.pos = pos
        self.reason = reason




from .grammar_parse_error import GRAMMAR_PARSER_ERROR_HEADER_MESSAGE, GrammarParseError


def describe_grammar_parse_error():
    def test_it_renders_a_message():
        grammar = "aa\\nbb\\ncc\\ndd\\nee"
        pos = 5
        reason = "reason"
        err = GrammarParseError("\n".join(grammar.split("\\n")), pos, reason)
        assert err.args[0] == "\n".join(
            [
                GRAMMAR_PARSER_ERROR_HEADER_MESSAGE(reason),
                "",
                "aa",
                "bb",
                "cc",
                " ^",
            ],
        )



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeGuard, TypeVar

from .grammar_graph_types import (
    RuleChar,
    RuleCharExclude,
    RuleEnd,
    UnresolvedRule,
)
from .rule_ref import RuleRef

if TYPE_CHECKING:
    from .graph_pointer import GraphPointer


T = TypeVar("T")


def type_guard(predicate: Callable[[Any], bool]) -> Callable[[Any], TypeGuard[T]]:
    def wrapper(value: Any) -> TypeGuard[T]:
        return predicate(value)

    return wrapper


@type_guard
def is_rule(rule: Any) -> bool:
    return rule is not None and isinstance(
        rule,
        (RuleChar | RuleCharExclude | RuleEnd | RuleRef),
    )


@type_guard
def is_rule_ref(rule: UnresolvedRule) -> bool:
    return isinstance(rule, RuleRef)


@type_guard
def is_rule_end(rule: UnresolvedRule) -> bool:
    return rule is not None and isinstance(rule, RuleEnd) and isinstance(rule, RuleEnd)


@type_guard
def is_rule_char(rule: UnresolvedRule) -> bool:
    return (
        rule is not None and isinstance(rule, RuleChar) and isinstance(rule, RuleChar)
    )


@type_guard
def is_rule_char_exclude(rule: UnresolvedRule) -> bool:
    return rule is not None and isinstance(rule, RuleCharExclude)


@type_guard
def is_range(inpt: Any) -> bool:
    return (
        isinstance(inpt, list)
        and len(inpt) == 2
        and all(isinstance(n, int) for n in inpt)
    )


@type_guard
def is_graph_pointer_rule_ref(pointer: GraphPointer) -> bool:
    return is_rule_ref(pointer.rule)


@type_guard
def is_graph_pointer_rule_end(pointer: GraphPointer) -> bool:
    return is_rule_end(pointer.rule)


@type_guard
def is_graph_pointer_rule_char(pointer: GraphPointer) -> bool:
    return is_rule_char(pointer.rule)


@type_guard
def is_graph_pointer_rule_char_exclude(pointer: GraphPointer) -> bool:
    return is_rule_char_exclude(pointer.rule)



import pytest

from .rule_ref import GraphNode, RuleRef


def describe_rule_ref():
    def test_initializes_with_a_given_value():
        ruleRef = RuleRef(123)
        assert ruleRef.value == 123

    def test_allows_setting_and_getting_nodes():
        mockNodes: set[GraphNode] = {1}
        ruleRef = RuleRef(123)
        ruleRef.nodes = mockNodes
        assert ruleRef.nodes == mockNodes

    def test_throws_an_error_if_trying_to_get_nodes_before_setting():
        ruleRef = RuleRef(123)
        with pytest.raises(ValueError, match="Nodes are not set") as context:
            _ = ruleRef.nodes
        assert str(context.value) == "Nodes are not set"



from .errors_types import ValidInput


def get_input_as_string(src: ValidInput) -> str:
    if isinstance(src, str):
        return src
    if isinstance(src, int):
        return chr(src)
    return "".join(chr(cp) for cp in src)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypedDict

from ..utils.validate_non_empty import validate_non_empty
from .rule_ref import RuleRef


class PrintOpts(TypedDict):
    pointers: set[GraphPointer]
    colorize: Callable[[str | int, str], str]
    show_position: bool


Range = tuple[int, int]


class Rule:
    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


@dataclass
class RuleWithValue(Rule):
    value: Any

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.value == other.value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(value={self.value})"


@dataclass
class RuleWithListOfIntsOrRanges(RuleWithValue):
    value: list[int | Range] = field(
        default_factory=list, metadata={"validate": validate_non_empty},
    )

    def __post_init__(self):
        self.value = self.value.copy()


@dataclass
class RuleChar(RuleWithListOfIntsOrRanges):
    pass


class RuleCharExclude(RuleWithListOfIntsOrRanges):
    pass


class RuleEnd(Rule):
    pass


UnresolvedRule = RuleChar | RuleCharExclude | RuleRef | RuleEnd


# ValidInput can either be a string, or a number indicating a code point.
# It CANNOT be a number representing a number; a number intended as input (like "8")
# should be passed in as a string.
ValidInput = str | int | list[int]


if TYPE_CHECKING:
    from .graph_pointer import GraphPointer

    # RuleRefs should never be exposed to the end user.
    ResolvedRule = RuleCharExclude | RuleChar | RuleEnd
    ResolvedGraphPointer = GraphPointer[ResolvedRule]
    Pointers = set[ResolvedGraphPointer]



from dataclasses import dataclass, field
from typing import Any

from ..utils.validate_non_empty import validate_non_empty


@dataclass
class InternalRuleDefWithNumericValue:
    value: int


@dataclass
class InternalBase:
    def __eq__(self, other):
        return isinstance(other, self.__class__)


@dataclass
class InternalBaseWithValue:
    value: Any

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.value == other.value


@dataclass
class InternalBaseWithInt(InternalBaseWithValue):
    value: int


@dataclass
class InternalBaseWithListOfInts(InternalBaseWithValue):
    value: list[int] = field(
        default_factory=list, metadata={"validate": validate_non_empty},
    )

    def __post_init__(self):
        self.value = self.value.copy()


@dataclass
class InternalRuleDefChar(InternalBaseWithListOfInts):
    pass


@dataclass
class InternalRuleDefCharAlt(InternalBaseWithInt):
    pass


class InternalRuleDefAlt(InternalBase):
    pass


@dataclass
class InternalRuleDefCharNot(InternalBaseWithListOfInts):
    pass


@dataclass
class InternalRuleDefCharRngUpper(InternalBaseWithInt):
    pass


@dataclass
class InternalRuleDefReference(InternalBaseWithInt):
    pass


@dataclass
class InternalRuleDefEnd(InternalBase):
    pass


@dataclass
class InternalRuleDefWithoutValue(InternalBase):
    pass


InternalRuleDef = (
    InternalRuleDefChar
    | InternalRuleDefEnd
    | InternalRuleDefReference
    | InternalRuleDefCharNot
    | InternalRuleDefWithNumericValue
    | InternalRuleDefWithoutValue
    | InternalRuleDefAlt
    | InternalRuleDefCharRngUpper
    | InternalRuleDefCharAlt
    | InternalRuleDefCharAlt
)

InternalRuleDefCharOrAltChar = InternalRuleDefChar | InternalRuleDefCharAlt

SymbolIds = dict[str, int]


def is_rule_def_alt(rule):
    return isinstance(rule, InternalRuleDefAlt)


def is_rule_def_ref(rule):
    return isinstance(rule, InternalRuleDefReference)


def is_rule_def_end(rule):
    return isinstance(rule, InternalRuleDefEnd)


def is_rule_def_char(rule):
    return isinstance(rule, InternalRuleDefChar)


def is_rule_def_char_not(rule):
    return isinstance(rule, InternalRuleDefCharNot)


def is_rule_def_char_alt(rule):
    return isinstance(rule, InternalRuleDefCharAlt)


def is_rule_def_char_rng_upper(rule):
    return isinstance(rule, InternalRuleDefCharRngUpper)



from .is_word_char import is_word_char


def test_should_return_true_for_lowercase_letters():
    assert is_word_char("a")
    assert is_word_char("z")


def test_should_return_true_for_uppercase_letters():
    assert is_word_char("A")
    assert is_word_char("Z")


def test_should_return_false_for_digits():
    assert not is_word_char("0")
    assert not is_word_char("9")


def test_should_return_false_for_non_word_characters():
    assert not is_word_char("-")
    assert not is_word_char("@")
    assert not is_word_char("_")
    assert not is_word_char("?")
    assert not is_word_char(" ")



import pytest

from ..utils.errors import GrammarParseError
from .parse_name import PARSE_NAME_ERROR, parse_name


def test_should_return_correct_name_when_encountering_a_valid_name():
    src = "validName"
    name = parse_name(src, 0)
    assert name == src


def test_should_return_correct_name_when_encountering_a_valid_name_starting_at_a_non_zero_position():
    src = "123validName"
    name = parse_name(src, 3)
    assert name == "validName"


def test_should_throw_error_when_encountering_an_invalid_name():
    src = "123"
    with pytest.raises(GrammarParseError) as exc_info:
        parse_name(src, 0)
    assert str(exc_info.value) == str(GrammarParseError(src, 0, PARSE_NAME_ERROR))



import pytest

from ..utils.errors import GrammarParseError
from .parse_char import parse_char


@pytest.mark.parametrize(
    ("description", "char", "code_point"),
    [
        ("escaped 8-bit unicode char", "a", ord("a")),
        ("escaped 8-bit unicode char", "9", ord("9")),
    ],
)
def test_parse_char_simple(description, char, code_point):
    grammar = f'root ::= "{char}" "foo"'
    assert parse_char(grammar, len('root ::= "')) == (code_point, 1)


@pytest.mark.parametrize(
    ("description", "escaped_char", "code_point", "inc_pos"),
    [
        ("escaped 8-bit unicode char", "\\x2A", ord("\x2A"), 4),
        ("escaped 16-bit unicode char", "\\u006F", ord("\u006F"), 6),
        ("escaped 32-bit unicode char", "\\U0001F4A9", 128169, 10),
        ("escaped tab char", "\\t", ord("\t"), 2),
        ("escaped new line char", "\\n", ord("\n"), 2),
        ("escaped \r char", "\\r", ord("\r"), 2),
        ("escaped quote char", '\\"', ord('"'), 2),
        ("escaped [ char", "\\[", ord("["), 2),
        ("escaped ] char", "\\]", ord("]"), 2),
        ("escaped \\ char", "\\\\", ord("\\"), 2),
    ],
)
def test_parse_char_complex(description, escaped_char, code_point, inc_pos):
    grammar = f'root ::= "{escaped_char}" "foo"'
    assert parse_char(grammar, len('root ::= "')) == (code_point, inc_pos)


@pytest.mark.parametrize(
    ("input", "pos"),
    [
        ("", 0),
        ("a", 1),
        ("a", 2),
    ],
)
def test_parse_char_raises(input, pos):
    with pytest.raises(GrammarParseError):
        parse_char(input, pos)



from __future__ import annotations

import pytest

from .build_error_position import build_error_position


def describe_build_error_position():
    @pytest.mark.parametrize(
        ("grammar", "pos", "expected"),
        [
            (
                'root ::= "foo"',
                1,
                [
                    'root ::= "foo"',
                    " ^",
                ],
            ),
            (
                'root ::= "foo"',
                5,
                [
                    'root ::= "foo"',
                    "     ^",
                ],
            ),
            # multi line grammars with pos on first line
            (
                "aa\\nbb",
                1,
                [
                    "aa",
                    " ^",
                ],
            ),
            # multi line grammars with pos on second line, first character
            (
                "aa\\nbb",
                2,
                [
                    "aa\\nbb",
                    "^",
                ],
            ),
            # multi line grammars with pos on second line, second character
            (
                "aa\\nbb",
                2 + 1,
                [
                    "aa\\nbb",
                    " ^",
                ],
            ),
            # multi line grammars beyond error with pos on second line
            (
                "aa\\nbb\\ncc",
                2 + 1,
                [
                    "aa\\nbb",
                    " ^",
                ],
            ),
            # multi line grammars beyond error with pos on third line, first char
            (
                "aa\\nbb\\ncc",
                2 + 2 + 0,
                [
                    "aa\\nbb\\ncc",
                    "^",
                ],
            ),
            # multi line grammars beyond error with pos on third line, second char
            (
                "aa\\nbb\\ncc",
                2 + 2 + 1,
                [
                    "aa\\nbb\\ncc",
                    " ^",
                ],
            ),
            # multi line grammars beyond error with pos on fourth line, first char
            (
                "aa\\nbb\\ncc\\ndd",
                2 + 2 + 2 + 0,
                [
                    "bb\\ncc\\ndd",
                    "^",
                ],
            ),
            # multi line grammars beyond error with pos on fifth line, second char
            (
                "aa\\nbb\\ncc\\ndd\\nee",
                2 + 2 + 2 + 2 + 1,
                [
                    "cc\\ndd\\nee",
                    " ^",
                ],
            ),
        ],
    )
    def it_correctly_shows_position_for_single_line_grammars(grammar, pos, expected):
        grammar_out, pos_out = expected
        result = build_error_position("\n".join(grammar.split("\\n")), pos)
        assert result == [
            *grammar_out.split("\\n"),
            pos_out,
        ]

    def test_it_renders_a_message_for_empty_input():
        assert build_error_position("", 0) == ["No input provided"]



from .parse_space import parse_space


def test_return_input_string_when_no_whitespace_or_comments():
    input_str = "abcdefghijk"
    pos = parse_space(input_str, 0, True)
    assert input_str[pos:] == input_str


def test_skip_leading_spaces_and_tabs():
    input_str = "   \t   abcdefghijk"
    expected = "abcdefghijk"
    pos = parse_space(input_str, 0, True)
    assert input_str[pos:] == expected


def test_skip_leading_newline_characters_when_newline_ok_true():
    input_str = "\n\n\r\n\r\nabcdefghijk"
    expected = "abcdefghijk"
    pos = parse_space(input_str, 0, True)
    assert input_str[pos:] == expected


def test_not_skip_leading_newline_characters_when_newline_ok_false():
    input_str = "\n\n\r\n\r\nabcdefghijk"
    expected = "\n\n\r\n\r\nabcdefghijk"
    pos = parse_space(input_str, 0, False)
    assert input_str[pos:] == expected


def test_skip_comments_and_leading_spaces():
    input_str = "  # This is a comment\n\t   abcdefghijk"
    expected = "abcdefghijk"
    pos = parse_space(input_str, 0, True)
    assert input_str[pos:] == expected


def test_skip_comments_and_leading_newline_characters_when_newline_ok_true():
    input_str = "\n\n # This is a comment\n\r\n\r\nabcdefghijk"
    expected = "abcdefghijk"
    pos = parse_space(input_str, 0, True)
    assert input_str[pos:] == expected


def test_skip_comments_and_leading_newline_characters_when_newline_ok_false():
    input_str = "\n\n # This is a comment\n\r\n\r\nabcdefghijk"
    expected = "\n\n # This is a comment\n\r\n\r\nabcdefghijk"
    pos = parse_space(input_str, 0, False)
    assert input_str[pos:] == expected


def test_return_empty_string_if_input_all_whitespace_and_comments():
    input_str = "  \t# Comment\n# Another comment\n\n"
    expected = ""
    pos = parse_space(input_str, 0, True)
    assert input_str[pos:] == expected


def test_return_empty_string_for_empty_input_string():
    input_str = ""
    expected = ""
    pos = parse_space(input_str, 0, True)
    assert input_str[pos:] == expected



from .input_parse_error import INPUT_PARSER_ERROR_HEADER_MESSAGE, InputParseError


def describe_input_parse_error():
    def test_it_renders_a_message():
        input = "some input"
        pos = 1
        err = InputParseError(input, pos)
        assert err.args[0] == "\n".join(
            [
                INPUT_PARSER_ERROR_HEADER_MESSAGE,
                "",
                input,
                " ^",
            ],
        )

    def test_it_renders_a_message_for_code_point():
        pos = 0
        err = InputParseError("a", pos)
        assert err.args[0] == "\n".join(
            [
                INPUT_PARSER_ERROR_HEADER_MESSAGE,
                "",
                "a",
                "^",
            ],
        )

    def test_it_renders_a_message_for_code_points():
        pos = 2
        err = InputParseError("abcd", pos)
        assert err.args[0] == "\n".join(
            [
                INPUT_PARSER_ERROR_HEADER_MESSAGE,
                "",
                "abcd",
                "  ^",
            ],
        )



from __future__ import annotations

from typing import Any, cast
from unittest.mock import MagicMock

from .grammar_graph_types import RuleChar
from .graph import Graph
from .graph_node import GraphNode
from .graph_pointer import GraphPointer

# if TYPE_CHECKING:
#     from .grammar_graph_types import Pointers
from .parse_state import ParseState


class MockGraph:
    def __init__(self):
        self.add = MagicMock(side_effect=lambda _input, pointers: pointers)

    @property
    def grammar(self):
        return "sample-grammar"


def make_mock_graph():
    return cast(Graph, MockGraph())


mock_pointers = cast(
    Any,
    # Pointers,
    {
        GraphPointer(
            GraphNode(
                RuleChar(value=[65]),
                meta={"stackId": 1, "pathId": 1, "stepId": 1},
            ),
        ),
    },
)


def describe_parse_state():
    def test_constructs_with_given_graph_and_pointers():
        parse_state = ParseState(make_mock_graph(), mock_pointers)
        assert isinstance(parse_state, ParseState)

    def test_returns_unique_rules_from_pointers():
        parse_state = ParseState(make_mock_graph(), mock_pointers)
        rules = list(parse_state.rules())
        assert len(rules) == 1
        assert isinstance(rules[0], RuleChar)
        assert rules[0].value == [65]

    def test_iterates_over_unique_rules_using_symbol_iterator():
        parse_state = ParseState(make_mock_graph(), mock_pointers)
        rules = list(parse_state.rules())
        assert len(rules) == 1
        assert isinstance(rules[0], RuleChar)

    def test_adds_new_input_and_returns_new_parse_state_with_updated_pointers():
        mock_graph = make_mock_graph()
        new_pointers = set({"rule": {"type": "char", "value": [66]}})
        mock_graph.add.mock_return_value(new_pointers)

        parse_state = ParseState(mock_graph, mock_pointers)
        new_state = parse_state.add("B")
        assert isinstance(new_state, ParseState)
        mock_graph.add.assert_called_with("B", mock_pointers)

    def test_calculates_the_size_of_unique_rules_correctly():
        mock_graph = make_mock_graph()
        parse_state = ParseState(mock_graph, mock_pointers)
        size = parse_state.size
        assert size == 1

    def test_provides_access_to_the_underlying_graph_grammar():
        mock_graph = make_mock_graph()
        parse_state = ParseState(mock_graph, mock_pointers)
        assert parse_state.grammar == "sample-grammar"

    def test_is_callable_as_a_function():
        mock_graph = make_mock_graph()
        parse_state = ParseState(mock_graph, mock_pointers)
        new_state = parse_state("B")
        assert isinstance(new_state, ParseState)

    def test_it_implements_adder_dunder_method():
        mock_graph = make_mock_graph()
        parse_state = ParseState(mock_graph, mock_pointers)
        new_state = parse_state + "B"
        assert isinstance(new_state, ParseState)



from __future__ import annotations

from .grammar_graph_types import RuleChar, UnresolvedRule
from .graph import Graph


def describe_graph():
    grammar = "example-grammar"
    stackedRules: list[list[list[UnresolvedRule]]] = [
        [
            [RuleChar(value=[65, 66, 67])],
            [RuleChar(value=[68, 69, 70])],
        ],
        [
            [RuleChar(value=[71, 72, 73])],
            [RuleChar(value=[74, 75, 76])],
        ],
    ]
    root_id = 0

    def test_should_create_a_graph_instance():
        graph = Graph(grammar, stackedRules, root_id)
        assert isinstance(graph, Graph)
        assert graph.grammar == grammar

    def test_should_get_the_root_node():
        graph = Graph(grammar, stackedRules, root_id)
        root_node = graph.get_root_node(root_id)
        assert isinstance(root_node, dict)
        assert len(root_node) == 2

    def test_should_get_the_initial_pointers():
        graph = Graph(grammar, stackedRules, root_id)
        root_node = graph.get_root_node(root_id)
        assert isinstance(root_node, dict)
        assert len(root_node) == 2

    def test_should_print_the_graph():
        graph = Graph(grammar, stackedRules, root_id)
        printed_graph = graph.print(colors=True)
        assert isinstance(printed_graph, str)
        assert len(printed_graph) > 0



from __future__ import annotations

from typing import Generic, TypedDict, TypeVar

from .grammar_graph_types import PrintOpts, UnresolvedRule
from .print import print_graph_node
from .rule_ref import RuleRef


class GraphNodeMeta(TypedDict):
    stackId: int
    pathId: int
    stepId: int


T = TypeVar("T", bound=UnresolvedRule)


class GraphNode(Generic[T]):
    rule: T
    next: GraphNode | None
    meta: GraphNodeMeta
    __id__: str | None = None

    def __init__(
        self,
        rule: T,
        meta: GraphNodeMeta | None,
        next_node: GraphNode | None = None,
    ):
        self.rule = rule
        if meta is None:
            raise ValueError("Meta is undefined")
        self.meta = meta
        self.next = next_node

    @property
    def id(self) -> str:
        if self.__id__ is None:
            self.__id__ = (
                f"{self.meta['stackId']},{self.meta['pathId']},{self.meta['stepId']}"
            )
        return self.__id__

    def print(self, opts: PrintOpts) -> str:
        return print_graph_node(self)(opts)


GraphNodeRuleRef = GraphNode[RuleRef]



from ..utils.errors import GrammarParseError
from .is_word_char import is_word_char

PARSE_NAME_ERROR = "Failed to find a valid name"

VALID_NAME_SEPARATORS = [
    "-",
    "_",
]


def parse_name(grammar: str, pos: int) -> str:
    name = ""
    while pos < len(grammar) and (
        is_word_char(grammar[pos]) or grammar[pos] in VALID_NAME_SEPARATORS
    ):
        name += grammar[pos]
        pos += 1
    if not name:
        raise GrammarParseError(grammar, pos, PARSE_NAME_ERROR)
    return name



from __future__ import annotations

from unittest.mock import patch

import pytest

from .grammar_graph_types import RuleChar
from .graph_node import GraphNode
from .print import print_graph_node, print_graph_pointer
from .rule_ref import RuleRef


@pytest.fixture(autouse=True)
def mock_get_parent_stack_id():
    with patch("gbnf.grammar_graph.print.get_parent_stack_id") as mock:
        yield mock


COLORS = {
    "\x1b[34m": "BLUE",
    "\x1b[36m": "CYAN",
    "\x1b[32m": "GREEN",
    "\x1b[31m": "RED",
    "\x1b[90m": "GRAY",
    "\x1b[33m": "YELLOW",
}


def mock_colorize(text, color):
    if color not in COLORS:
        raise ValueError(f"Invalid color: {color}")
    return f"[{COLORS[color]}]:{text}"


class MockNode:
    def __init__(self, id, rule, next: GraphNode | None = None):
        self.id = id
        self.rule = rule
        self.next = next
        self.print = lambda _opts: f"Node({id})"


def create_mock_node(id, rule, next: GraphNode | None = None):
    return MockNode(id, rule, next)


def create_mock_graph_pointer(node: MockNode):
    return {
        "node": node,
        "parent": None,
        "print": lambda _opts: f"Pointer to {node.id}",
    }


def describe_print():
    def describe_print_graph_pointer():
        def test_it_prints_graph_pointer_details_correctly(mock_get_parent_stack_id):
            mock_node = create_mock_node(1, RuleRef(100))
            mock_pointer = create_mock_graph_pointer(mock_node)
            mock_get_parent_stack_id.return_value = "foo"
            result = print_graph_pointer(mock_pointer)(
                {"colorize": mock_colorize, "pointers": set(), "show_position": False},
            )
            assert result == "[RED]:*foo"

    def describe_print_graph_node():
        def test_it_prints_graph_node_with_a_character_rule():
            rule_char = RuleChar(value=[65])
            mock_node = create_mock_node(1, rule_char)
            result = print_graph_node(mock_node)(
                {"colorize": mock_colorize, "show_position": False, "pointers": set()},
            )
            assert result == "[GRAY]:[[YELLOW]:A[GRAY]:]"

        def test_it_prints_graph_node_with_a_rule_reference():
            rule_ref = RuleRef(200)
            mock_node = create_mock_node(1, rule_ref)
            result = print_graph_node(mock_node)(
                {"colorize": mock_colorize, "show_position": True, "pointers": set()},
            )
            assert result == "[BLUE]:{[GRAY]:1[BLUE]:}[GRAY]:Ref([GREEN]:200[GRAY]:)"
